@@ -24,16 +24,20 @@ $, = " "; $\ = $/;
 my $t1 = 5;    		# number of
 my $t2 = 5;    		# tests per
 my $t3 = 2;         # section
+my $t4 = 3;
+
+# $t1 = $t2 = $t3 = 0;
 
 use Cwd; $CWD = cwd();
+$SRV = "$CWD/echosrv";
 eval "use File::Temp qw/tempdir/";
 $t3 = 0 if $@;
 
-eval "use Test::Simple tests => 4 + $t1 + $t2 + $t3";
+eval "use Test::Simple tests => 4 + $t1 + $t2 + $t3 + $t4";
 
 my $debug = $ENV{DEBUG} || 0;
-#sub POE::Kernel::TRACE_GARBAGE ()  { 1 }
-#sub POE::Kernel::ASSERT_DEFAULT () { 1 }
+sub POE::Kernel::TRACE_GARBAGE ()  { 1 }
+sub POE::Kernel::ASSERT_DEFAULT () { 1 }
 
 use POE qw(Component::Child Filter::Stream);
 ok(1, 'use PoCo::Child'); # If we made it this far, we're ok.
@@ -56,20 +60,13 @@ my %t3 = (
     stdout  => "t3_out",
     );
 
-# create main session
-
-$r = POE::Session->create(
-    package_states => [
-        main => [ values(%t1), values(%t2), values(%t3) ]
-        ],
-    inline_states => {
-    	_start => sub { $_[KERNEL]->alias_set("main"); },
-    	_stop => sub { print "_stop" if $debug; },
-    	_default => \&_default,
-    	}
+my %t4 = (
+    stderr  => "t4_error",
+    died    => "t4_died",
+    done    => "t4_done",
     );
 
-ok(defined($r), "session created");
+my %t = reverse(%t1, %t2, %t3, %t4);
 
 # test non-interactive child
 
@@ -77,8 +74,8 @@ if ($t1) {
     $t1 = POE::Component::Child->new(
     	events => \%t1, debug => $debug
     	);
-    ok(defined $t1 && $t1->isa('POE::Component::Child'), "component 1");
-    $t1->run("$^X echosrv --stdout");
+    ok(defined $t1 && $t1->isa('POE::Component::Child'), "[1] init - non-interactive");
+    $t1->run($^X, $SRV, "--stdout");
     }
 
 # test interactive child
@@ -89,9 +86,8 @@ if ($t2) {
     	events => { %t2, done => \&t2_done },
     	debug => $debug
     	);
-    ok(defined $t2 && $t2->isa('POE::Component::Child'), "component 2");
-    $t2->run("$^X echosrv");
-    $t2->write("hej");
+    ok(defined $t2 && $t2->isa('POE::Component::Child'), "[2] init - interactive");
+    $t2->run($^X, $SRV);
     }
 
 # test other stuff
@@ -101,10 +97,37 @@ if ($t3) {
     $t3 = POE::Component::Child->new(
     	events => \%t3, debug => $debug, chdir => $TMPDIR,
     	);
-    ok(defined $t3 && $t3->isa('POE::Component::Child'), "component 3");
+    ok(defined $t3 && $t3->isa('POE::Component::Child'), "[3] init - other");
     $t3->run("pwd");
     }
 
+if ($t4) {
+    $t4 = POE::Component::Child->new(
+    	events => \%t4, debug => $debug
+    	);
+    ok(defined $t4 && $t4->isa('POE::Component::Child'), "[4] init - pipe close");
+    $t4->run($^X, $SRV);
+    }
+
+# create main session
+
+$r = POE::Session->create(
+    package_states => [
+        main => [ keys %t ]
+        ],
+    inline_states => {
+    	_start => sub {
+            $_[KERNEL]->alias_set("main");
+            $t2->write("hej");
+            $t4->write('hello');
+            },
+    	_stop => sub { print "_stop" if $debug; },
+    	_default => \&_default,
+    	}
+    );
+
+ok(defined($r), "session created");
+
 # POEtry in motion
 
 POE::Kernel->run();
@@ -118,73 +141,88 @@ sub t1_out {
     my ($self, $args) = @_[ARG0 .. $#_];
     local $_ = $args->{out};
 
-    ok(/echo/, "standard output");
+    ok(/server/, "[1]: standard output");
     }
 
 sub t1_err {
     my ($self, $args) = @_[ARG0 .. $#_];
-    ok($args->{out} =~ /echo/, "standard error");
+    my $err = $args->{out} =~ /server/;
+    ok($err, $err ? "[1]: standard error" : "[1]: $args->{out}");
     }
 
-$t1n = 0;
+BEGIN { $t1n = 0; }
 sub t1_done {
     my ($self, $args) = @_[ARG0 .. $#_];
 
     if ($t1n++ == 0) {
-    	$t1->run("$^X echosrv", "--stderr");
+    	$t1->run($^X, $SRV, "--stderr");
     	}
 
     else {
-    	ok(1, "done event");
-    	$t1->run("$^X echosrv", "--die");
+    	ok(1, "[1]: done event");
+    	$t1->run($^X, $SRV, "--die");
     	}
     }
 
 sub t1_died {
-    ok(1, "died tested");
+    ok(1, "[1]: died tested");
     }
 
 sub t1_error {
     my ($self, $args) = @_[ARG0 .. $#_];
-    ok(0, "Co1: unexpected error: $args->{error}");
+    ok(0, "[1]: unexpected error: $args->{error}");
     }
 
 # --- event handlers - interactive child --------------------------------------
 
 sub t2_err {
     my ($self, $args) = @_[ARG0 .. $#_];
-    ok($args->{out} eq "hej", "client write tested");
+    ok($args->{out} eq "hej", "[2]: client write tested");
     $t2->quit();
     }
 
-$t2n = 0;
+BEGIN { $t2n = 0; }
 sub t2_done {
     my ($self, $args) = @_[ARG0 .. $#_];
 
     if ($t2n++) {
-    	ok(0, "kill problem");
+    	ok(0, "[2]: kill problem");
     	return;
     	}
 
-    ok(1, "callback references");
-    ok(1, "quit method");
-    $t2->run("$^X echosrv");
+    ok(1, "[2]: callback references");
+    ok(1, "[2]: quit method");
+    $t2->run($^X, $SRV);
     $t2->kill();
     }
 
 sub t2_died {
     my ($kernel, $self, $args) = @_[KERNEL, ARG0 .. $#_];
-    ok(1, "killed");
+    ok(1, "[2]: killed");
     }
     
 sub t2_error {
     my ($self, $args) = @_[ARG0 .. $#_];
-    ok(0, "Co2: unexpected error: $args->{error}");
+    ok(0, "[2]: unexpected error: $args->{error}");
     }
 
 sub t3_out {
     my ($self, $args) = @_[ARG0 .. $#_];
-    ok($args->{out} eq $TMPDIR, "dir change");
+    ok($args->{out} eq $TMPDIR, "[3]: dir change");
+    }
+
+sub t4_error {
+    my ($self, $args) = @_[ARG0 .. $#_];
+    ok($args->{out} eq "hello", "[4]: child talks");
+    $self->quit();
+    }
+
+sub t4_died {
+    ok(1, "[4]: stdout closed");
+    }
+
+sub t4_done {
+    ok(1, "[5]: done");
     }
 
 sub _default {
